<?php
/**
 * @version SOFORT Gateway 4.2.0 - $Date: 2012-09-06 13:49:09 +0200 (Thu, 06 Sep 2012) $
 * @author SOFORT AG (integration@sofort.com)
 * @link http://www.sofort.com/
 *
 * Copyright (c) 2012 SOFORT AG
 *
 * $Id: helperFunctions.php 5326 2012-09-06 11:49:09Z boehm $
 * 
 * This file includes functions used by module and by callback.php
 *
 */

class HelperFunctions {
	
	
	/**
	 * Fill table sofort_orders with a new order
	 * @param int $ordersId
	 * @param string $paymentSecret
	 * @param string $transactionId
	 * @param string $paymentMethod
	 * @return last insert_id ELSE data could not be saved
	 */
	public static function insertSofortOrder($ordersId, $paymentSecret, $transactionId, $paymentMethod) {
		global $db;
		
		if(!$transactionId || !$paymentMethod) return false;
		
		switch($paymentMethod){
			case 'SR': $paymentMethod = 'rechnung_by_sofort'; break;
			case 'SU': $paymentMethod = 'sofortueberweisung'; break;
			case 'SV': $paymentMethod = 'vorkasse_by_sofort'; break;
			case 'SL': $paymentMethod = 'sofortlastschrift'; break;
			case 'LS': $paymentMethod = 'lastschrift_by_sofort'; break;
			default: return false;
		}
		
		$sqlDataArray = array(
				'orders_id' => $ordersId,
				'transaction_id' => $transactionId,
				'payment_method' => $paymentMethod,
				'payment_secret' => $paymentSecret,
		);
		$db->Execute(HelperFunctions::getEscapedInsertInto(HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders', $sqlDataArray));
		return $db->insert_ID(); // fetch and return the last insert id
	}
	
	
	/**
	 * Fill table sofort_orders_notification, e.g. in case of status-changes - each notification = one entry
	 * @see updateTimeline()
	 * @param transactionID
	 * @param - array $dbEntries - array with key=dbFieldName and value=dbFieldValue (all dbFieldNames are optional) (id and date_time will be ignored)
	 * @return int - last insert_id
	 */
	public static function insertSofortOrdersNotification($transactionId, $dbEntries, $SofortLoggingObj = '') {
		global $db;
		
		$sofortOrdersId 	= (isset($dbEntries['sofort_orders_id'])) 	? $dbEntries['sofort_orders_id'] 	: '';
		$items 				= (isset($dbEntries['items'])) 				? $dbEntries['items'] 				: '';
		$amount 			= (isset($dbEntries['amount'])) 			? $dbEntries['amount'] 				: '';
		$customerComment	= (isset($dbEntries['customer_comment'])) 	? $dbEntries['customer_comment'] 	: '';
		$sellerComment 		= (isset($dbEntries['seller_comment'])) 	? $dbEntries['seller_comment'] 		: '';
		$statusId 			= (isset($dbEntries['status_id'])) 			? $dbEntries['status_id'] 			: '';
		$status 			= (isset($dbEntries['status'])) 			? $dbEntries['status'] 				: '';
		$statusReason 		= (isset($dbEntries['status_reason'])) 		? $dbEntries['status_reason'] 		: '';
		$invoiceStatus 		= (isset($dbEntries['invoice_status'])) 	? $dbEntries['invoice_status'] 		: '';
		$invoiceObjection 	= (isset($dbEntries['invoice_objection'])) 	? $dbEntries['invoice_objection'] 	: '';
		
		if (!$sofortOrdersId) {
			$result = $db->Execute('SELECT id FROM '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders WHERE transaction_id = "'.HelperFunctions::escapeSql($transactionId).'"');
			$sofortOrdersId = $result->fields['id'];
			if (!$sofortOrdersId) {
				if (is_object($SofortLoggingObj)) {
					$SofortLoggingObj->logWarning("Notification could not be inserted into table sofort_notifications. Transaction-ID: $transactionId, Data:".print_r($dbEntries, true)." Order-ID: $orderId Check order and order-history.");
				}
				return false;
			}
		}
		
		$sqlDataArray = array(
			'sofort_orders_id' => $sofortOrdersId,
			'items' => $items,
			'amount' => $amount,
			'customer_comment' => $customerComment,
			'seller_comment' => $sellerComment,
			'status_id' => $statusId,
			'status' => $status,
			'status_reason' => $statusReason,
			'invoice_status' => $invoiceStatus,
			'invoice_objection' => $invoiceObjection,
		);

		$db->Execute(HelperFunctions::getEscapedInsertInto(HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders_notification', $sqlDataArray));
		return $db->insert_ID(); // fetch and return the last insert id
	}
	
	
	/**
	 * return the last/penultimate value of the given field
	 * @param int $ordersId
	 * @param string $field
	 * @param bool $returnPenultimateEntry (optional) - Default = false
	 */
	public static function getLastFieldValueFromSofortTable($ordersId, $field, $returnPenultimateEntry = false){
		global $db;
		
		$result = $db->Execute('SELECT id FROM '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders WHERE orders_id = '.HelperFunctions::escapeSql($ordersId));
		$sofortOrdersId = $result->fields['id'];
		if ($returnPenultimateEntry) {
			$result = $db->Execute('
				SELECT '.HelperFunctions::escapeSql($field).' 
				FROM '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders_notification 
				WHERE sofort_orders_id = "'.HelperFunctions::escapeSql($sofortOrdersId).'" 
				ORDER BY date_time DESC 
				LIMIT 1,1');
		} else {
			$result = $db->Execute('
				SELECT '.HelperFunctions::escapeSql($field).'
				FROM '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders_notification 
				WHERE sofort_orders_id = "'.HelperFunctions::escapeSql($sofortOrdersId).'" 
				ORDER BY date_time DESC 
				LIMIT 1');
		}
		return $result->fields[$field];
	}
	
	
	/**
	 * Converts a given String $string to any specified encoding (if supported)
	 *
	 * @param String $string
	 * @param String $to ; 2 = from utf-8 to shopencoding set in sofort.ini ; 3 = from shopencoding set in sofort.ini to utf-8
	 * @return String $string
	 */
	public static function convertEncoding($string, $to, $fromEncoding = '') {
		$shopEncoding = HelperFunctions::getIniValue('shopEncoding');
		
		if ($shopEncoding == 'UTF-8'){
			return $string;
		} elseif ($to == 1) {
			return mb_convert_encoding($string, $shopEncoding, $fromEncoding);
		} elseif ($to == 2) {
			return mb_convert_encoding($string, $shopEncoding, 'UTF-8');
		} elseif ($to == 3){
			return mb_convert_encoding($string, 'UTF-8', $shopEncoding);
		}
	}
	
	
	/**
	 * escapes the given string via mysql_real_esacpe_string (if function exists & a db-connection is available) or mysql_escape_string
	 * @param string $string
	 * @return string $string
	 */
	public static function escapeSql($string) {
		if (function_exists('mysql_real_escape_string') && mysql_ping()) {
			return mysql_real_escape_string($string);
		} else {
			return mysql_escape_string($string);
		}
	}
	
	
	/**
	 * Combination of functions escapeSql() and convertEncoding()
	 */
	public static function escapeConvert($string, $to) {
		return HelperFunctions::escapeSql(HelperFunctions::convertEncoding($string,$to));
	}
	
	
	/**
	 * creates an escaped "INSERT INTO" sql-string
	 * @param string $table - complete name of table (incl. DB-prefix, if set)
	 * @param array $data with key=column_name and value=column_value
	 * for sql-commands set value like "sqlcommand:now()"
	 * @return string $returnString
	 */
	public static function getEscapedInsertInto($table, $data) {
		$table = trim($table);
		
		if (!is_string($table) || !$table) {
			return '';
		}
		
		if (!is_array($data) || !$data) {
			return '';
		}
		
		$returnString = 'INSERT INTO `'.HelperFunctions::escapeSql($table).'` (`';
		$columns = array_keys($data);
		$returnString .= implode('`, `', $columns);
		$returnString .= '`) VALUES (';
		
		foreach ($data as $value) {
			if ((strpos($value, 'sqlcommand:') === 0)) {
				$returnString .= HelperFunctions::escapeSql(substr($value, 11)).", ";  //its a sql-command
			}else{
				$returnString .= "'".HelperFunctions::escapeSql($value)."', ";  //its a normal string or int
			}
		}
		
		$returnString = substr($returnString, 0, -2); //deletes comma and whitespace
		$returnString .= ')'; //dont add ';'
		return $returnString;
	}
	
	
	/**
	 * masks the given variable via strip_tags() and htmlentities()
	 * @param mixed $var
	 * @return mixed $var
	 */
	public static function htmlMask($var){
		return htmlentities(strip_tags($var),ENT_QUOTES,HelperFunctions::getIniValue('shopEncoding'));
	}
	
	
	/**
	 * return from the sofort.ini the value of the given key
	 * @return mixed $value - if key is not found: false
	 */
	public static function getIniValue($key){
		$iniArray = parse_ini_file('sofort.ini');
		if (!isset($iniArray[$key])) {
			return false;
		}else{
			return $iniArray[$key];
		}
	}
	
	
	public static function getSofortLanguage($language){
		$lngdir = DIR_WS_LANGUAGES;
		$lngarr = array();
		
		foreach (new DirectoryIterator($lngdir) as $file){
			if (file_exists($lngdir.$file->getFilename().'/modules/payment/sofort_general.php')) $lngarr[] = $file->getFilename();
		}
		
		return (!in_array($language,$lngarr)) ? 'english' : $language;
	}
	
	
	public static function getShortCode($lng){
		switch ($lng){
			case 'german' : return 'de';
			case 'dutch'  : return 'nl';
			case 'french' : return 'fr';
			case 'italian': return 'it';
			case 'polish' : return 'pl';
			case 'english':
			default		  : return 'en';
		}
	}
	
	
	/**
	 * Getter for the sofortmodulVersion, set in sofort.ini
	 * @return String
	 */
	public static function getSofortmodulVersion() {
		return HelperFunctions::getIniValue('sofortmodulVersion');
	}
	
	
	/**
	 * validate given api-key against SOFORT and return result
	 * @return bool
	 */
	public static function apiKeyIsValid($apiKey) {
		preg_match('#([a-zA-Z0-9:]+)#', $apiKey, $matches);
		$configKey = $matches[1];
		$SofortLib_TransactionData = new SofortLib_TransactionData($configKey);
		$SofortLib_TransactionData->setTransaction('00000')->sendRequest();
		if ($SofortLib_TransactionData->isError()) {
			return false;
		} else {
			return true;
		}
	}
	
	
	/**
	 * get the cancel-URL - add given errors to the url
	 * @param $paymentMethod - assign $this->code
	 * @param array (optional) $errors with error-codes (structure like XML-structure!)
	 * @param string $cancelUrl (optional) - if given use this error-url than standard-error-url
	 */
	public static function getCancelUrl($paymentMethod, $errors = array(), $cancelUrl = '') {
		
		$params = 'payment_error='.$paymentMethod;
		
		if ($errors) {
			$errorCodes = array();
			foreach ($errors as $oneError) {
				$errorCodes[] = $oneError['code'];
			}
			if($errorCodes){
				$params .= '&error_codes='.implode(',', $errorCodes);
			}
		}
		
		if (!$cancelUrl) {
			return zen_href_link(FILENAME_CHECKOUT_PAYMENT, $params, 'SSL', true, false);
		} else {
			return $cancelUrl .= $params;
		}
		
	}
	
	
	/**
	 * return the last orderStatus for the given orderId from table orders (NOT from sofort-tables!)
	 */
	public static function getLastOrderStatus($orderId){
		global $db;
		if (!$orderId)
			return false;
		
		$result = $db->Execute('SELECT orders_status 
				FROM '.HelperFunctions::escapeSql(TABLE_ORDERS).' 
				WHERE orders_id = "'.HelperFunctions::escapeSql($orderId).'"');
		return $result->fields['orders_status'];
	}
	
	
	/**
	 * check, if given paymentSecret belongs to the given transId
	 * @param string $paymentSecretToCheck
	 * @param string $transactionId
	 * @return bool
	 */
	public static function checkPaymentSecret($paymentSecretToCheck, $transactionId) {
		global $db;
		$result = $db->Execute('SELECT payment_secret FROM '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders WHERE transaction_id = "'.HelperFunctions::escapeSql($transactionId).'"');
		$dbPaymentSecret = $result->fields['payment_secret'];
		if ($paymentSecretToCheck == $dbPaymentSecret) {
			return true;
		} else {
			return false;
		}
	}
	
	
	/**
	 * trim the given params to needed shorter string e.g.: 'sr', 'pending', 'confirm_invoice' --> "sr_pen_con_inv"
	 * @return string
	 */
	public static function getShortStatus ($paymentMethod, $status, $statusReason) {
		$longStatus = $paymentMethod.'_'.$status.'_'.$statusReason;
		$parts = explode('_', $longStatus);
	
		foreach ($parts as $key => $part) {
			$part = trim($part);
			$parts[$key] = substr($part, 0, 3);
		}
	
		$parts = implode('_', $parts);
		return $parts;
	}
	
	
	/**
	 * get for the given notification-status the history-comments for the customer and seller
	 * @param bool $checkAgainstPenultimateEntry - if for current notification the entry into sofort_orders_notification was done, all checks have to be done against the penultimate entry
	 * @return array - Comments for the customer and seller (they can also be empty or only one of them is set!)
	 */
	public static function getHistoryComments($notificationStatusShort, $orderId, $SofortLib_TransactionData, $transactionId, $checkAgainstPenultimateEntry = false, $SofortLoggingObj = ''){
		$historyComments = array();
		$historyComments['seller'] = '';
		$historyComments['buyer'] = '';
		
		$supportedStatus = array(
				'SV_PEN_WAI_FOR_MON', 'SV_REC_CRE', 'SV_REC_CON_PRO', 'SV_REC_OVE', 'SV_REC_PAR_CRE', 'SV_REF_COM', 'SV_REF_REF', 'SV_LOS_NOT_CRE',
				'SU_PEN_NOT_CRE_YET', 'SU_LOS_NOT_CRE', 'SU_REC_CRE', 'SU_REF_REF', 'SU_REF_COM',
				'LS_REC_CRE', 'LS_LOS_REJ', 'LS_REF_COM', 'LS_REF_REF', 'LS_PEN_NOT_CRE_YET',
				'SL_REC_CRE', 'SL_LOS_REJ', 'SL_REF_COM', 'SL_REF_REF', 'SL_PEN_NOT_CRE_YET',
				'SR_PEN_CON_INV', 'SR_LOS_CAN', 'SR_PEN_NOT_CRE_YET', 'SR_REF_REF', 'SR_LOS_CON_PER_EXP',
		);
		
		if (!in_array(strtoupper($notificationStatusShort), $supportedStatus)) {
			if (is_object($SofortLoggingObj)) {
				$SofortLoggingObj->log("Paymentstatus ($notificationStatusShort) is unimportant. No comments will be set to history. Order-ID: $orderId");
			}
			//normally we could exit() here but maybe there has been a change of the SR-amount
			return $historyComments;
		}
		
		//prevent identical comments
		if (HelperFunctions::statusAndReasonDidNotChange($orderId, $SofortLib_TransactionData->getStatus(), $SofortLib_TransactionData->getStatusReason(), $checkAgainstPenultimateEntry)) {
			if (is_object($SofortLoggingObj)) {
				$SofortLoggingObj->log("Paymentstatus ($notificationStatusShort) did not change since last notification. No comments will be set to history. Maybe comments have been inserted by successUrl/notification earlier! This notice is no fault! Order-ID: $orderId");
			}
			return $historyComments;
		}
		
		if (defined('MODULE_PAYMENT_SOFORT_'.strtoupper($notificationStatusShort).'_BUYER')) {
			$historyComments['buyer'] = constant('MODULE_PAYMENT_SOFORT_'.strtoupper($notificationStatusShort).'_BUYER');
		}
		
		if (defined('MODULE_PAYMENT_SOFORT_'.strtoupper($notificationStatusShort).'_SELLER')) {
			$historyComments['seller'] = constant('MODULE_PAYMENT_SOFORT_'.strtoupper($notificationStatusShort).'_SELLER');
		}
		
		//onlyest status where the transId has to be added is sr_pen_con_inv
		if ($notificationStatusShort == 'sr_pen_con_inv' && !empty($historyComments['seller']) && $transactionId) {
			$historyComments['seller'] .= $transactionId;
		}
		
		//special replacements in history-strings
		$historyComments = HelperFunctions::replaceSpecialTextInHistoryComment($historyComments, $SofortLib_TransactionData, $transactionId);
		
		return $historyComments;
	}
	
	
	/**
	 * get the new shop-status for the given status
	 * @param string $notificationStatus - e.g. "sr_pen_con_inv"
	 * @param int $orderId
	 * @param object $SofortLoggingObj - only for logging of notices/warnings/errors and other messages
	 * @return int newShopStatus for this notification OR false if status-update is unimportant (e.g. there could be an amount-update needed)
	 */
	public static function getNewShopStatusId($notificationStatusShort, $orderId, $SofortLoggingObj = '') {
		$statusConstant = strtoupper('MODULE_PAYMENT_SOFORT_'.$notificationStatusShort.'_STATUS_ID');
		
		//exceptions
		switch ($statusConstant) {
			case 'MODULE_PAYMENT_SOFORT_SR_LOS_CON_PER_EXP_STATUS_ID':
				$statusConstant = 'MODULE_PAYMENT_SOFORT_SR_LOS_CAN_STATUS_ID';
				break;
			case 'MODULE_PAYMENT_SOFORT_SV_REC_OVE_STATUS_ID':
			case 'MODULE_PAYMENT_SOFORT_SV_REC_PAR_CRE_STATUS_ID':
				$statusConstant = 'MODULE_PAYMENT_SOFORT_SV_WRONG_AMOUNT_STATUS_ID';
				break;
		}
		
		//the needed status-constant must be defined otherwise status is unimportant (or there is an error e.g. if seller has uninstalled the module)
		if (defined($statusConstant)) {
			$newShopStatus = constant($statusConstant);
		} else {
			if (is_object($SofortLoggingObj)) {
				$SofortLoggingObj->log("Required status ($statusConstant) not defined/found. Status and statusupdate is unimportant! Order-ID: $orderId! (This might be an error, if module was not correct installed or is currently uninstalled!)");
			}
			//we should but cannot exit here, there could be a change of the SR-Amount with an unimportant status
			return false;
		}
		
		//dont set a new status? -> then use the last order status
		if (HelperFunctions::statusIsUnchangedStatus($newShopStatus)) {
			return HelperFunctions::getLastOrderStatus($orderId);
		}else{
			return $newShopStatus;
		}
	}
	
	
	/**
	 * check if given status is "dont change - use last order status" ( = MODULE_PAYMENT_SOFORT_MULTIPAY_UNCHANGED_STATUS_ID)
	 * @param int $statusToCheck
	 * @param object $SofortLoggingObj (optional)
	 * @return bool
	 */
	public static function statusIsUnchangedStatus($statusToCheck, $SofortLoggingObj = '') {
		if (!defined('MODULE_PAYMENT_SOFORT_MULTIPAY_UNCHANGED_STATUS_ID')) {
			if (is_object($SofortLoggingObj)) {
				$SofortLoggingObj->logWarning("No 'unchanged'-Status (MODULE_PAYMENT_SOFORT_MULTIPAY_UNCHANGED_STATUS_ID) found. Maybe module was not correct installed. Please read the manual! Unchanged-Status will be ignored. Check order! Order-ID: $orderId");
			}
			return false;
		}
		
		if ($statusToCheck == MODULE_PAYMENT_SOFORT_MULTIPAY_UNCHANGED_STATUS_ID) {
			return true;
		} else {
			return false;
		}
	}
	
	
	/**
	 * check if last notification had the same status and statusReason - e.g. SR: pending-not_credited_yet_pending is the same like pending-not_credited_yet_delcredere
	 * @return bool
	 */
	public static function statusAndReasonDidNotChange($orderId, $statusToCheck, $statusReasonToCheck, $checkAgainstPenultimateEntry = false) {
		$lastStatus = HelperFunctions::getLastFieldValueFromSofortTable($orderId, 'status', $checkAgainstPenultimateEntry);
		$lastStatusReason = HelperFunctions::getLastFieldValueFromSofortTable($orderId, 'status_reason', $checkAgainstPenultimateEntry);
		
		if ($statusToCheck == $lastStatus && $statusReasonToCheck == $lastStatusReason) {
			return true;
		} else {
			return false;
		}
	}
	
	
	/**
	 * replace special parts of the history-comment e.g. "{{time}}" -> "01.01.2000 10:10", "{{amount_refunded}}" -> "123" etc.
	 * @param array $historyComments within seller- and buyer-comment
	 * @param object $SofortLib_TransactionData
	 * @param int $transactionId
	 * @return array $historyComment within all needed replacements
	 */
	public static function replaceSpecialTextInHistoryComment($historyComments, $SofortLib_TransactionData, $transactionId){
		//used in regex below, dont change the naming!
		$refunded_amount = $SofortLib_TransactionData->getAmountRefunded();
		$tId = $transactionId;
		$transaction = $transactionId;
		$paymentMethod = $SofortLib_TransactionData->getPaymentMethod();
		$time = date('d.m.Y, G:i:s');
		
		$paymentMethodStr = '';
		$ks = ($SofortLib_TransactionData->getConsumerProtection()) ? 'KS_' : '';
		$paymentMethodStr = constant('MODULE_PAYMENT_SOFORT_'.strtoupper($paymentMethod).'_'.$ks.'TEXT_TITLE');
		
		//some string may have inside {{time}} - its not needed anymore
		$historyComments['seller'] = trim(str_ireplace('{{time}}', '', $historyComments['seller']));
		$historyComments['buyer']  = trim(str_ireplace('{{time}}', '', $historyComments['buyer']));
		
		//replacement
		$historyComments['seller'] = preg_replace('#\{\{([a-zA-Z0-9_]+)\}\}#e', '$$1', $historyComments['seller']);
		$historyComments['buyer']  = preg_replace('#\{\{([a-zA-Z0-9_]+)\}\}#e', '$$1', $historyComments['buyer']);
		
		return $historyComments;
	}
	
	
	/**
	 * Fetch order-ID from table orders for given transactionId
	 * @param $transactionId
	 * @return int order-ID ELSE empty string
	 */
	public static function getOrderId($transactionId, $SofortLoggingObj = '') {
		global $db;
		$result = $db->Execute('SELECT orders_id FROM '.HelperFunctions::escapeSql(TABLE_ORDERS_STATUS_HISTORY).' WHERE comments = "'.HelperFunctions::escapeSql($transactionId).'" LIMIT 1');
		$orderId = $result->fields['orders_id'];
		if (empty($orderId)) {
			if (is_object($SofortLoggingObj)) {
				$SofortLoggingObj->logWarning("Order-ID not found for Transaction-ID: $transactionId. Check order!");
			}
			return '';
		}
		
		return $orderId;
	}
	
	
	/**
	 * (only SR) check given notification-amount against the last shop-amount
	 * @param int $notificationAmount
	 * @param int $orderId
	 * @param bool $usePenultimateEntry (optional) - Standard=false - Check given amount against the penultimate entry in sofort_orders_notification
	 * @param object $SofortLoggingObj (optional)
	 * @return bool true if amount differ ELSE false
	 */
	public static function checkIfNewTotal($notificationAmount, $orderId, $usePenultimateEntry = false, $SofortLoggingObj = ''){
		$lastShopTotal = HelperFunctions::getLastFieldValueFromSofortTable($orderId, 'amount', $usePenultimateEntry);
		
		//1st notification? -> amount cannot be changed
		if (!$lastShopTotal) {
			return false;
		}
		
		//warning, if no notification exists in sofort-table
		if($lastShopTotal === '') {
			if (is_object($SofortLoggingObj)) {
				$SofortLoggingObj->logWarning("Last total (amount) could not be found in shop-DB! Check amount of the order! Order-ID: $orderId");
			}
			return false;
		}
		
		if ($lastShopTotal != $notificationAmount){
			return true;
		} else {
			return false;
		}
	}
	
	
	/**
	 * Only SR: insert a comment into buyer-history with a text like "Invoice changed. New amount: 123 Euro."
	 * @param int $orderId
	 * @param int $status
	 * @param string $time (optional) - will be placed in comment if given
	 * @param float $oldTotal
	 * @param float $newTotal
	 * @param object $SofortLoggingObj (optional)
	 * @return bool true (always)
	 */
	public static function insertNewTotalCommentToHistory($orderId, $status, $oldTotal, $newTotal, $time = '', $SofortLoggingObj = ''){
		global $db;
		if ($newTotal > $oldTotal){
			$comments = MODULE_PAYMENT_SOFORT_SR_TRANSLATE_CART_RESET.' '.MODULE_PAYMENT_SOFORT_SR_TRANSLATE_CURRENT_TOTAL.' '.$newTotal.' Euro';
		} else {
			$comments = MODULE_PAYMENT_SOFORT_SR_TRANSLATE_CART_EDITED.' '.MODULE_PAYMENT_SOFORT_SR_TRANSLATE_CURRENT_TOTAL.' '.$newTotal.' Euro';
		}
		
		if ($time) {
			$comments .= ' '.MODULE_PAYMENT_SOFORT_TRANSLATE_TIME.': '.$time;
		}
		
		HelperFunctions::insertHistoryEntry($orderId, $status, $comments);
		
		if (is_object($SofortLoggingObj)) {
			$SofortLoggingObj->log("New total comment set to history. Old total: $oldTotal, new total: $newTotal, Order-ID: ".$orderId);
		}
		return true;
	}
	
	
	/**
	 * update delivery- and invoice-address in shopsystem, if buyer changed them at pnag-wizard
	 * @return bool true (always)
	 */
	public static function updateShopAdresses ($invoiceAddress, $shippingAddress, $orderId, $SofortLoggingObj = '') {
		global $db;
		
		$dbResult = $db->Execute("SELECT countries_name FROM ".HelperFunctions::escapeSql(TABLE_COUNTRIES)." WHERE countries_iso_code_2 = '".HelperFunctions::escapeConvert($invoiceAddress['country_code'],2)."'");
		$invoiceCountry = ($dbResult->fields['countries_name']) ? $dbResult->fields['countries_name'] : $invoiceAddress['country_code'];
		$sql = "UPDATE	".HelperFunctions::escapeSql(TABLE_ORDERS)."
				SET		billing_name = '" .HelperFunctions::escapeConvert($invoiceAddress['firstname'],2)." ".HelperFunctions::escapeConvert($invoiceAddress['lastname'],2)."', 
						billing_street_address = '".HelperFunctions::escapeConvert($invoiceAddress['street'],2)." ".HelperFunctions::escapeConvert($invoiceAddress['street_number'],2)."',
						billing_suburb = '".HelperFunctions::escapeConvert($invoiceAddress['street_additive'],2)."',
						billing_city = '".HelperFunctions::escapeConvert($invoiceAddress['city'],2)."',
						billing_postcode = '".HelperFunctions::escapeConvert($invoiceAddress['zipcode'],2)."',  
						billing_company = '',
						billing_state = '',
						billing_country = '".HelperFunctions::escapeConvert($invoiceCountry,2)."', 
						last_modified = now() 
				WHERE	orders_id = '".(int)$orderId."'";
		
		$db->Execute($sql);
		if (is_object($SofortLoggingObj)) {
			if (function_exists('mysql_affected_rows')) {
				switch (mysql_affected_rows()) {
					case 1: 
						$SofortLoggingObj->log("Invoiceaddress successfully updated in shop. Order-ID: $orderId.");
						break;
					case 0:
						$SofortLoggingObj->log("Invoiceaddress did not change. Update not necessary in Shop. Maybe address was updates earlier by successUrl/Notification. Order-ID: $orderId.");
						break;
					case -1:
						//currently nothing to do here
				}
			} 
		}
		
		$dbResult = $db->Execute("SELECT countries_name FROM ".HelperFunctions::escapeSql(TABLE_COUNTRIES)." WHERE countries_iso_code_2 = '".HelperFunctions::escapeConvert($shippingAddress['country_code'],2)."'");
		$shippingCountry = ($dbResult->fields['countries_name']) ? $dbResult->fields['countries_name'] : $shippingAddress['country_code'];
		$sql = "UPDATE ".HelperFunctions::escapeSql(TABLE_ORDERS)."
				SET delivery_name = '".HelperFunctions::escapeConvert($shippingAddress['firstname'],2)." ".HelperFunctions::escapeConvert($shippingAddress['lastname'],2)."',
					delivery_street_address = '".HelperFunctions::escapeConvert($shippingAddress['street'],2)." ".HelperFunctions::escapeConvert($shippingAddress['street_number'],2)."',
					delivery_suburb = '".HelperFunctions::escapeConvert($shippingAddress['street_additive'],2)."',
					delivery_city = '".HelperFunctions::escapeConvert($shippingAddress['city'],2)."',
					delivery_postcode = '".HelperFunctions::escapeConvert($shippingAddress['zipcode'],2)."',
					delivery_company = '',
					delivery_state = '', 
					delivery_country = '".HelperFunctions::escapeConvert($shippingCountry,2)."',
					last_modified = now()
				WHERE	orders_id = '".(int)$orderId."'";
		
		$db->Execute($sql);
		if (is_object($SofortLoggingObj)) {
			if (function_exists('mysql_affected_rows')) {
				switch (mysql_affected_rows()) {
					case 1: 
						$SofortLoggingObj->log("Shippingaddress successfully updated in shop. Order-ID: $orderId.");
						break;
					case 0:
						$SofortLoggingObj->log("Shippingaddress did not change. Update not necessary in Shop. Maybe address was updates earlier by successUrl/Notification. Order-ID: $orderId.");
						break;
					case -1:
						//currently nothing to do here
				}
			} 
		}
		
		return true;
	}
	
	
	/**
	 * set a new payment-status to an order in table "orders"
	 * @param int $orderId
	 * @param int $newStatus
	 * @return bool true (always)
	 */
	public static function updateTableOrdersStatus($orderId, $newStatus) {
		global $db;
		$db->Execute('UPDATE '.HelperFunctions::escapeSql(TABLE_ORDERS).' SET orders_status = "'.HelperFunctions::escapeSql($newStatus).'", last_modified = now() WHERE orders_id = '.(int)$orderId.';');
		return true;
	}
	
	
	/**
	 * set new history-entry into table "orders_status_history"
	 * @param int $orderId
	 * @param int $status - if '-1' -> comment is only shown to seller (not buyer)
	 * @param string $comment for buyer/seller
	 * @param string $time (optional) - will be placed behind the comment
	 * @return bool true (always)
	 */
	public static function insertHistoryEntry($orderId, $status, $comment, $customerNotified = '', $time = '') {
		global $db;
		if ($time) {
			$comment .= ' '.MODULE_PAYMENT_SOFORT_TRANSLATE_TIME.': '.$time;
		}
		
		if (!is_numeric($customerNotified)) {
			$customerNotified = 0;
		}
		
		$sqlDataArray = array(
			'orders_id'			=> (int)$orderId,
			'orders_status_id'	=> $status,
			'date_added'		=> 'sqlcommand:now()',
			'customer_notified' => (int) $customerNotified,
			'comments'			=> $comment,
		);
		$db->Execute(HelperFunctions::getEscapedInsertInto(TABLE_ORDERS_STATUS_HISTORY, $sqlDataArray));
		return true;
	}
	
	
	/**
	 * check, if there is at least one entry in table sofort_orders_notification for given transactionId
	 * @return bool
	 */
	public static function hasEntriesInTableSofortNotifications($transactionId) {
		global $db;
		$result = $db->Execute('
			SELECT '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders.id 
			FROM `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders_notification`
			LEFT JOIN `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders` 
			ON '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders.id = '.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders_notification.sofort_orders_id
			WHERE transaction_id = "'.HelperFunctions::escapeSql($transactionId).'"');
		
		if ($result->RecordCount() > 0) {
			return true;
		} else {
			return false;
		}
	}
	
	
	/**
	 * get transaction-ID for given order-ID from table_orders 
	 * @return int $transactionID (might be empty, if not found)
	 */
	public static function getTransactionIdFromShop($orderId){
		global $db;
		$result = $db->Execute('
			SELECT h.comments
			FROM '.HelperFunctions::escapeSql(TABLE_ORDERS_STATUS_HISTORY).' AS h, '.HelperFunctions::escapeSql(TABLE_ORDERS).' AS o
			WHERE o.orders_id = h.orders_id 
			AND o.orders_id = "'.(int) $orderId.'"
			AND h.comments != "" 
			AND comments REGEXP "^[0-9]{5}-[0-9]{5,6}-[0-9A-F]{8}-[0-9A-F]{4}$"
			LIMIT 1');
		return $result->fields['comments'];
	}
	
	
	/**
	 * replace '&amp;' AND '&&' into '&' --- zen_href_link() doesnt work always correct --- ATTENTION: check, if it is needed for your url-string!
	 * @param string $stringToClean - complete url or only parameters
	 * @return string - cleaned string
	 */
	public static function cleanUrlParameter($stringToClean) {
		$stringToClean = str_replace('&amp;', '&', $stringToClean);
		$stringToClean = str_replace('&&', '&', $stringToClean);
		
		return $stringToClean;
	}
	
	
	/**
	 * sort the SOFORT-item-ID by value
	 * example: "49{1}26{6}23" and "49{6}23{1}26" is the same article -> sort by number "...{1}...{6}..." and return it
	 * option-fields like "13_chk47" will be observed
	 */
	public static function sortSofortItemId($sofortItemId) {
		if (strpos($sofortItemId, '{') === false) {
			return $sofortItemId; //article has no attributes, return article-id
		}
		
		$splittet = explode('{', $sofortItemId);
		$newSofortItemId = $splittet[0]; //first comes article-ID
		
		unset ($splittet[0]); //now there are only attributes inside
		
		$sortedAttributes = array();
		foreach ($splittet as $optionAndValue) {
			$explodedOptionAndValue = explode('}', $optionAndValue);
			
			//Zen-Bug: option-ID can be in Frontend "13_chk47" and in backend "13" --> we always set "13"
			//if there are more "features" like this in the article-attributes -> add/replace here
			preg_match('/^[0-9]*/', $explodedOptionAndValue[0], $explodedOptionAndValue[0]);
			
			$sortedAttributes[$explodedOptionAndValue[0][0]] = $explodedOptionAndValue[1]; //keys are unique
		}
		
		ksort($sortedAttributes);
		
		foreach ($sortedAttributes as $option => $value) {
			$newSofortItemId .= '{'.$option.'}'.$value;
		}
		
		return $newSofortItemId; //return something like "49{1}26{6}23"
	}
}