<?php
/**
 * @version SOFORT Gateway - $Date: 2012-11-09 10:17:24 +0100 (Fr, 09 Nov 2012) $
 * @author SOFORT AG (integration@sofort.com)
 * @link http://www.sofort.com/
 *
 * Copyright (c) 2012 SOFORT AG
 *
 * $Id: sofort.php 5634 2012-11-09 09:17:24Z rotsch $
 */

require_once ((IS_ADMIN_FLAG === true ? DIR_FS_CATALOG_MODULES : DIR_WS_MODULES).'payment/sofort/library/sofortLib.php');
require_once(dirname(__FILE__).'/helperFunctions.php');

/**
 * Superclass of ZenCart modules
 */
class sofort {
	
	var $transaction_id, $auth_code, $email_footer, $order_status, $enabled, $code, $description, $title, $sort_order; //shop-vars
	var $sofort_js, $sofort, $invoice, $sofortPaymentUrl, $paymentMethod, $title_extern; //sofort-vars
	
	/**
	 * checked ok
	 */
	function sofort(){
		if (!isset($_SESSION['sofort'])) {
			$_SESSION['sofort'] = array();
		}
		
		//shop-vars
		$this->email_footer = ''; //used by shopsystem - currently not in use
		$this->transaction_id = ''; //shop-var, exists for preventing php-notice
		$this->auth_code = ''; //shop-var, exists for preventing php-notice
		$this->order_status = (defined('MODULE_PAYMENT_SOFORT_MULTIPAY_TEMP_STATUS_ID') ? MODULE_PAYMENT_SOFORT_MULTIPAY_TEMP_STATUS_ID : (int)DEFAULT_ORDERS_STATUS_ID);
		//sofort-vars
		$this->sofort_js = false;
		$this->sofort = '';
		$this->invoice = '';
		$this->sofortPaymentUrl = '';
		//other class-vars are set in child
	}
	
	
	/**
	 * checked ok
	 * check if payment method is allowed in the payment zone and payment country
	 * if not: module will be disabled
	 */
	function update_status() {
		global $order, $db;
		
		$constantName = 'MODULE_PAYMENT_SOFORT_'.$this->paymentMethod.'_ZONE';
		
		if (($this->enabled == true) && ((int)constant($constantName) > 0)) {
			$check_flag = false;
			$check = $db->Execute("SELECT	zone_id
								   FROM		".HelperFunctions::escapeSql(TABLE_ZONES_TO_GEO_ZONES)."
								   WHERE	geo_zone_id = '".HelperFunctions::escapeSql(constant($constantName))."'
								   AND		zone_country_id = '".(int)$order->billing['country']['id']."'
								   ORDER	by zone_id");
			
			while (!$check->EOF) {
				if ($check->fields['zone_id'] < 1) {
					$check_flag = true;
					break;
				} elseif ($check->fields['zone_id'] == $order->billing['zone_id']) {
					$check_flag = true;
					break;
				}
				
				$check->MoveNext();
			}
			
			if ($check_flag == false) $this->enabled = false;
		}
		
		$constantName = 'MODULE_PAYMENT_SOFORT_'.$this->paymentMethod.'_COUNTRIES';
		$constantValue = constant($constantName);
		$constantValue = trim($constantValue);
		$country_zones = explode(",", $constantValue);
		
		foreach ($country_zones as &$country_zone) {
			$country_zone = strtoupper($country_zone);
		}
		
		if($this->enabled != false && $constantValue) {
			$dest_country = $order->billing['country']['iso_code_2'];
			
			foreach ($country_zones as &$country_zone) {
				$country_zone = trim ($country_zone);
			}
			
			if (!in_array($dest_country, $country_zones)) $this->enabled = false;
		}
	}
	
	
	/**
	 * checked ok
	 */
	function javascript_validation() {
		return '';
	}
	
	
	/**
	 * checked ok
	 * extended in all modules "by sofort"
	 */
	function pre_confirmation_check() {
		//prevent loops, because before_process() is called twice for each order
		//we only allow to checkout, if buyer pass this code here
		$_SESSION['sofort']['checkout_allowed'] = 1;
	}
	
	
	/**
	 * checked ok
	 * call with parent::selection() in child
	 */
	function selection() {
		//$this->_checkCancelOrder(); //dont activate in Zen! May result in loop!
		
		//return false, if modul was not correct updated
		if (!defined('MODULE_PAYMENT_SOFORT_MULTIPAY_MODULE_VERSION') ||
				strcasecmp(trim(MODULE_PAYMENT_SOFORT_MULTIPAY_MODULE_VERSION), trim(HelperFunctions::getSofortmodulVersion())) != 0) {
			if (is_object($this->sofort)) {
				$this->sofort->logWarning("Paymentmodul was updated by seller but not correct installed. Installed-version in DB does not match file-versions. Please take a look at MODULE_PAYMENT_SOFORT_MULTIPAY_UPDATE_NOTICE or read the manual!");
			} elseif (is_object($this->invoice)) {
				$this->invoice->logWarning("Paymentmodul was updated by seller but not correct installed. Installed-version in DB does not match file-versions. Please take a look at MODULE_PAYMENT_SOFORT_MULTIPAY_UPDATE_NOTICE or read the manual!");
			}
			return false;
		}
		
		if (isset($_SESSION['sofort']['apiKeyIsValid'])) {
			return $_SESSION['sofort']['apiKeyIsValid'];
		} else if (!isset($_SESSION['sofort']['apiKeyIsValid'])) {
			$apiTestResult = HelperFunctions::apiKeyIsValid(MODULE_PAYMENT_SOFORT_MULTIPAY_APIKEY);
			if (!$apiTestResult) {
				if (is_object($this->sofort)) {
					$this->sofort->log("Notice: Function apiKeyIsValid() returned false.");
				} elseif (is_object($this->invoice)) {
					$this->invoice->log("Notice: Function apiKeyIsValid() returned false.");
				}
			}
			$_SESSION['sofort']['apiKeyIsValid'] = $apiTestResult;
			return $apiTestResult;
		}
	}
	
	
	/**
	 * checked ok
	 */
	function confirmation() {
		
		$this->_checkCancelOrder();
		
	}
	
	
	/**
	 * checked ok
	 * does nothing
	 */
	function process_button() {
	}
	
	
	/**
	 * This function is called twice! Before AND after payment at SOFORT
	 */
	function before_process() {
		//if buyer has successfully payed, delete all session-data and ALWAYS redirect to shop-success-page
		if (isset($_REQUEST['sofort_transaction']) && isset($_SESSION['sofort']['sofort_transaction']) &&
				$_REQUEST['sofort_transaction'] == $_SESSION['sofort']['sofort_transaction'] &&
				isset($_REQUEST['success_secret']) && isset($_SESSION['sofort']['success_secret']) &&
				$_REQUEST['success_secret'] == $_SESSION['sofort']['success_secret']) {
			$orderId = substr($_SESSION['sofort']['cart_pn_sofortueberweisung_id'], strpos($_SESSION['sofort']['cart_pn_sofortueberweisung_id'], '-') + 1);
			$transactionId = $_SESSION['sofort']['sofort_transaction'];
			
			//delete sofort-variables
			if (isset($_SESSION['sofort']['sofort_ls_sender_holder'])) 		unset ($_SESSION['sofort']['sofort_ls_sender_holder']);
			if (isset($_SESSION['sofort']['sofort_ls_account_number']))		unset ($_SESSION['sofort']['sofort_ls_account_number']);
			if (isset($_SESSION['sofort']['sofort_ls_bank_code'])) 			unset ($_SESSION['sofort']['sofort_ls_bank_code']);
			if (isset($_SESSION['sofort']['sofort_conditions_sr'])) 		unset ($_SESSION['sofort']['sofort_conditions_sr']);
			if (isset($_SESSION['sofort']['sofort_conditions_sv'])) 		unset ($_SESSION['sofort']['sofort_conditions_sv']);
			if (isset($_SESSION['sofort']['sofort_conditions_ls'])) 		unset ($_SESSION['sofort']['sofort_conditions_ls']);
			if (isset($_SESSION['sofort']['sofort_transaction'])) 			unset ($_SESSION['sofort']['sofort_transaction']);
			if (isset($_SESSION['sofort']['cart_pn_sofortueberweisung_id']))unset ($_SESSION['sofort']['cart_pn_sofortueberweisung_id']);
			if (isset($_SESSION['sofort']['success_secret']))				unset ($_SESSION['sofort']['success_secret']);
			if (isset($_SESSION['sofort']['sofort_transaction']))			unset ($_SESSION['sofort']['sofort_transaction']);
			
			$_SESSION['sofort']['checkout_allowed'] = 0;
			
			//delete shop-session-data
			$this->_resetCartAndDeleteSessionData($orderId, $transactionId);
			
			//insert payed-status
			$this->_insertPaidStatus($orderId, $transactionId);
			
			if ($this->code == 'sofort_sv') {
				$this->_sendSvBankDataEmail();
				zen_redirect($this->_insertSvBankdataAndGetLinkToBankdataPage($orderId));
			} else {
				zen_redirect(zen_href_link(FILENAME_CHECKOUT_SUCCESS, '', 'SSL'));
			}
		}
		
		//for security for preventing payment-loops
		//dont insert order into db, if checkout is not allowed
		if ($_SESSION['sofort']['checkout_allowed'] != 1) {
			if (is_object($this->sofort)) {
				$this->sofort->logError("Error: Modul tries to go to checkout but is not allowed! Buyer will be redirected to payment-page. Order was not saved in shop-DB. Check status of orders (if exist) made at this time. Customer: ".$_SESSION['customer_id']);
			} elseif (is_object($this->invoice)) {
				$this->invoice->logError("Error: Modul tries to go to checkout but is not allowed! Buyer will be redirected to payment-page. Order was not saved in shop-DB. Check status of orders (if exist) made at this time. Customer: ".$_SESSION['customer_id']);
			}
			$cancelUrl = zen_href_link(FILENAME_CHECKOUT_PAYMENT, 'payment_error='.$this->code.'&error_codes=10008', 'SSL', true, false);
			$cancelUrl = HelperFunctions::cleanUrlParameter($cancelUrl);
			zen_redirect($cancelUrl);
		}
		
		//now, shopsystem will insert the order into table orders...
		return true;
	}
	
	
	/**
	 * call the SOFORT-API - called by shopsystem directly after saving of cart in shop-db
	 * overwritten by sr
	 * @param int $orderId
	 */
	function after_order_create($orderId) {
		global $order, $order_totals, $db;
		
		$customerId = $_SESSION['customer_id'];
		$currency = $order->info['currency'];
		
		//if buyer will not successfully abort the payment (e.g. close the SOFORT-Wizard) and the goes to the
		//shop-index.php and want to pay again -> set the orderStatus to "aborted"
		$_SESSION['sofort']['cart_pn_sofortueberweisung_id'] = $_SESSION['cart']->cartID . '-' . $orderId;
		
		$reasons = $this->_getReasons($this->paymentMethod, $customerId, $order, $orderId);
		
		$paymentSecret = md5(mt_rand().microtime()); //security for call of notification-url
		$successSecret = md5(mt_rand().microtime().time()); //security for call of success-url - will be checked, if success-url is called
		$_SESSION['sofort']['success_secret'] = $successSecret;
		
		$userVariable_0 = $orderId;
		$userVariable_1 = $customerId;
		$userVariable_2 = $paymentSecret;
		
		$amount = $this->_getShopTotal($order_totals, $order);
		
		$successUrl = zen_href_link(FILENAME_CHECKOUT_PROCESS, 'sofort_transaction=-TRANSACTION-&success_secret='.$successSecret, 'SSL', true, false);
		$successUrl = HelperFunctions::cleanUrlParameter($successUrl);
		
		$cancelUrl = zen_href_link(FILENAME_CHECKOUT_PAYMENT, 'sofort_transaction=-TRANSACTION-&payment_error='.$this->code, 'SSL', true, false);
		$cancelUrl = HelperFunctions::cleanUrlParameter($cancelUrl);
		
		$notificationUrl = zen_href_link('sofort_callback', 'paymentSecret='.$paymentSecret.'&sofort_action=multipay', 'SSL', false, false);
		$notificationUrl = HelperFunctions::cleanUrlParameter($notificationUrl);
		
		$this->sofort->setAmount($amount, $currency);
		$this->sofort->setReason(HelperFunctions::convertEncoding($reasons[0],3), HelperFunctions::convertEncoding($reasons[1],3));
		$this->sofort->setSuccessUrl(HelperFunctions::convertEncoding($successUrl,3));
		$this->sofort->setAbortUrl(HelperFunctions::convertEncoding($cancelUrl,3));
		$this->sofort->setTimeoutUrl(HelperFunctions::convertEncoding($cancelUrl,3));
		$this->sofort->setNotificationUrl(HelperFunctions::convertEncoding($notificationUrl,3));
		$this->sofort->addUserVariable(HelperFunctions::convertEncoding($userVariable_0,3));
		$this->sofort->addUserVariable(HelperFunctions::convertEncoding($userVariable_1,3));
		$this->sofort->addUserVariable(HelperFunctions::convertEncoding($userVariable_2,3));
		$this->sofort->setEmailCustomer(HelperFunctions::convertEncoding($order->customer['email_address'],3));
		$this->sofort->setPhoneNumberCustomer(HelperFunctions::convertEncoding($order->customer['telephone'],3));
		
		//set special settings for every payment-method
		switch($this->paymentMethod) {
			case 'SU' :
				$this->sofort->setSofortueberweisung();
				// see if customer protection is enabled, set it as parameter to sofortlib
				$this->sofort->setSofortueberweisungCustomerprotection(MODULE_PAYMENT_SOFORT_SU_KS_STATUS == 'True');
				break;
			case 'SL' :
				$this->sofort->setSofortlastschrift();
				$this->sofort->setSenderAccount(
					'',
					'',
					HelperFunctions::convertEncoding($order->billing['firstname'],3) . ' ' . HelperFunctions::convertEncoding($order->billing['lastname'],3)
				);
				break;
			case 'LS' :
				$this->sofort->setLastschrift();
				$this->sofort->setSenderAccount(
					HelperFunctions::convertEncoding($_SESSION['sofort']['sofort_ls_bank_code'],3),
					HelperFunctions::convertEncoding($_SESSION['sofort']['sofort_ls_account_number'],3),
					HelperFunctions::convertEncoding($_SESSION['sofort']['sofort_ls_sender_holder'],3)
				);
				
				$billingSalutation = $this->_getGenderFromAddressBook(
					$order->billing['firstname'],
					$order->billing['lastname'],
					$order->billing['company'],
					$order->billing['street_address'],
					$order->billing['postcode'],
					$order->billing['city'],
					$order->billing['country_id'],
					$order->billing['zone_id']
				);
				
				//split street and number
				if (!preg_match('#(.+)[ .](.+)#i', trim($order->billing['street_address']), $streetparts)) {
					$streetparts = array();
					$streetparts[1] = trim($order->billing['street_address']);
					$streetparts[2] = '';
				}
				
				$this->sofort->setLastschriftAddress(
					HelperFunctions::convertEncoding($order->billing['firstname'],3),
					HelperFunctions::convertEncoding($order->billing['lastname'],3),
					HelperFunctions::convertEncoding($streetparts[1],3),
					HelperFunctions::convertEncoding($streetparts[2],3),
					HelperFunctions::convertEncoding($order->billing['postcode'],3),
					HelperFunctions::convertEncoding($order->billing['city'],3),
					HelperFunctions::convertEncoding($billingSalutation,3),
					HelperFunctions::convertEncoding($order->billing['country']['iso_code_2'],3)
				);
				break;
			case 'SV' :
				$this->sofort->setSofortvorkasse();
				// if this is called a 'test transaction', add a sender account
				if (getenv('test_sv') == true) {
					$this->sofort->setSenderAccount('00000', '12345', 'Tester Testaccount');
				}
				//notice: customer-protection with SV is currently not supported and not shown to seller for activation
				//@see function keys()
				$this->sofort->setSofortvorkasseCustomerprotection(MODULE_PAYMENT_SOFORT_SV_KS_STATUS == 'True');
				break;
		}
		//send all data against the SOFORT-API
		$this->sofort->sendRequest();
		
		if($this->sofort->isError()) {
			$this->sofort->logWarning("API-Call returned false. Redirect to cancel-URL. API-Errors: ".print_r($this->sofort->getErrors(), true)." Time: ".date("d.m.Y, G:i:s"));
			zen_redirect(HelperFunctions::getCancelUrl($this->code, $this->sofort->getErrors()));
		} else {
			$url = $this->sofort->getPaymentUrl();
			$transactionId = $this->sofort->getTransactionId();
			
			//seller and customer comment
			$time = date("d.m.Y, G:i:s");
			
			//set temp-status (only table orders_history, not table orders)
			$tmpStatusId = MODULE_PAYMENT_SOFORT_MULTIPAY_TEMP_STATUS_ID;
			//dont set a new status? -> then use the last order status
			if (HelperFunctions::statusIsUnchangedStatus($tmpStatusId)) {
				$tmpStatusId = HelperFunctions::getLastOrderStatus($orderId);
			}
			
			//comment only for buyer and seller
			$tmpComment = constant('MODULE_PAYMENT_SOFORT_'.$this->paymentMethod.'_TMP_COMMENT');
			HelperFunctions::insertHistoryEntry($orderId, $tmpStatusId, $tmpComment, '', $time);
			
			HelperFunctions::insertHistoryEntry($orderId, -1, $transactionId);
			
			//save all important data in our sofort-tables
			$sofortOrderId = HelperFunctions::insertSofortOrder($orderId, $paymentSecret, $transactionId, $this->paymentMethod);
			if (!$sofortOrderId) {
				$this->sofort->logWarning("Warning: Saving of orderdetails in table sofort_orders failed. Function insertSofortOrder() returned false.
					Given params: OrderId: $orderId, PaymentSecret: $paymentSecret, TransId: $transactionId, PaymentMethod:".$this->paymentMethod);
			}
			
			$_SESSION['sofort']['sofort_transaction'] = $transactionId;
			
			//redirect to sofortPaymentUrl will be done later in after_process()
			$this->sofortPaymentUrl = $url;
		}
	}
	
	
	/**
	 * make redirect to sofortPaymentUrl - in case fo failures redirect to payment-page
	 * @return nothing - always redirects (to SOFORT-payment-URL or to error-URL)
	 */
	function after_process() {
		global $insert_id;
		//first security-checks
		if (!isset($this->sofortPaymentUrl) || !$this->sofortPaymentUrl){
			if (is_object($this->sofort)) {
				$this->sofort->logError("Error: Payment-URL must be, but is not set! Redirect customer to payment-page. Check status of order! Order-ID: ".$insert_id." Customer: ".$_SESSION['customer_id']);
			} elseif (is_object($this->invoice)) {
				$this->invoice->logError("Error: Payment-URL must be, but is not set! Redirect customer to payment-page. Check status of order! Order-ID: ".$insert_id." Customer: ".$_SESSION['customer_id']);
			}
			zen_redirect(HelperFunctions::getCancelUrl($this->code));
		} else {
			$sofortHost = (getenv('sofortApiUrl') != '') ? getenv('sofortApiUrl') : 'https://www.sofort.com';
			$hostToCheck = parse_url($sofortHost, PHP_URL_HOST);
			$paymentHost = parse_url($this->sofortPaymentUrl, PHP_URL_HOST);
			
			if (strpos($paymentHost, $hostToCheck) === false) {
				if (is_object($this->sofort)) {
					$this->sofort->logError("Error: Redirect buyer to SOFORT-Payment-URL failed because found host ($paymentHost) doesnt equal allowed host ($sofortHost). Order was already saved in shop-db. Order-ID: ".$insert_id." Customer: ".$_SESSION['customer_id']);
				} elseif (is_object($this->invoice)) {
					$this->invoice->logError("Error: Redirect buyer to SOFORT-Payment-URL failed because found host ($paymentHost) doesnt equal allowed host ($sofortHost). Order was already saved in shop-db. Order-ID: ".$insert_id." Customer: ".$_SESSION['customer_id']);
				}
				zen_redirect(HelperFunctions::getCancelUrl($this->code));
			}
		}
		
		//for security for preventing payment-loops
		if ($_SESSION['sofort']['checkout_allowed'] != 1) {
			if (is_object($this->sofort)) {
				$this->sofort->logError("Error: Modul tries to redirect to payment-url but is not allowed! Buyer will be redirected to payment-page. Order was already saved in shop-DB. Check status of orders. Order-ID: ".$insert_id." Customer: ".$_SESSION['customer_id']);
			} elseif (is_object($this->invoice)) {
				$this->invoice->logError("Error: Modul tries to redirect to payment-url but is not allowed! Buyer will be redirected to payment-page. Order was already saved in shop-DB. Check status of orders. Order-ID: ".$insert_id." Customer: ".$_SESSION['customer_id']);
			}
			$cancelUrl = zen_href_link(FILENAME_CHECKOUT_PAYMENT, 'payment_error='.$this->code.'&error_codes=10009', 'SSL', true, false);
			$cancelUrl = HelperFunctions::cleanUrlParameter($cancelUrl);
			zen_redirect($cancelUrl);
		}
		
		$_SESSION['sofort']['checkout_allowed'] = 0;
		
		//comment only for seller
		$time = date("d.m.Y, G:i:s");
		$tmpCommentSeller = constant('MODULE_PAYMENT_SOFORT_'.$this->paymentMethod.'_TMP_COMMENT_SELLER');
		HelperFunctions::insertHistoryEntry($insert_id, -1, $tmpCommentSeller, 0, $time);
		
		//everything ok - redirect to sofortPaymentPage
		zen_redirect($this->sofortPaymentUrl);
	}
	
	
	function get_error() {
		
		$this->_checkCancelOrder();
		
		//if no error occured or the code doesnt belong to this paymentMethod
		if (!isset($_GET['payment_error']) || $_GET['payment_error'] != $this->code) {
			return false;
		}
		
		$errormsgArray = array();
		
		if (isset($_GET['error_codes'])) {
			$langConstantExist = true;
			$errorCodes = array_unique (explode(',', $_GET['error_codes']));
			
			foreach ($errorCodes as $errorCode) {
				$errorCode = trim($errorCode);
				$code = substr($errorCode, 0, strpos($errorCode, '.'));
				
				if ($code === false || empty($code)) {
					$code = $errorCode;
				}
				
				if (defined('MODULE_PAYMENT_SOFORT_MULTIPAY_XML_FAULT_'.strtoupper($errorCode))) {
					$errormsgArray[] = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_XML_FAULT_'.strtoupper($errorCode)).' ('.$code.')';
				} else if(defined('MODULE_PAYMENT_SOFORT_MULTIPAY_XML_FAULT_'.$code)) {
					$errormsgArray[] = constant ('MODULE_PAYMENT_SOFORT_MULTIPAY_XML_FAULT_'.strtoupper($code)).' ('.$code.')';
				} else {
					//specific errorinfo not defined
					$langConstantExist = false;
				}
				
			}
			
			//disable module in view, if there is an error with SR (but dont disable, if conditions have not been checked)
			if ($this->code == "sofort_sr" && !in_array('10002', $errorCodes)) {
				$this->enabled = false;
			}
			
			if (!$errormsgArray && $langConstantExist == false) {
				$errormsgArray[] = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_XML_FAULT_'.$this->paymentMethod);
			}
		} else {
			$errormsgArray[] = constant('MODULE_PAYMENT_SOFORT_'.$this->paymentMethod.'_TEXT_ERROR_MESSAGE');
			
			if ($this->code == "sofort_sr") {
				$this->enabled = false;
			}
		}
		
		if (!$errormsgArray) {
			$errormsgArray[] = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_XML_FAULT_'.$this->paymentMethod);
		}
		
		$errormsgArray = array_unique($errormsgArray);
		
		$title = MODULE_PAYMENT_SOFORT_MULTIPAY_TEXT_ERROR_HEADING;
		return array('title' => $title,
			'error' => implode(', ', $errormsgArray));
	}
	
	
	/**
	 * checked ok
	 */
	function check() {
		global $db;
		
		if (!isset($this->_check)) {
			$constantName = 'MODULE_PAYMENT_SOFORT_'.$this->paymentMethod.'_STATUS';
			$check_query = $db->Execute('SELECT configuration_value FROM '.HelperFunctions::escapeSql(TABLE_CONFIGURATION).' WHERE configuration_key = "'.HelperFunctions::escapeSql($constantName).'"');
			$this->_check = $check_query->RecordCount();
		}
		return $this->_check;
	}
	
	
	/**
	 * checked ok
	 * install shared keys, that are used by all/most multipay-modules
	 * called by module with parent::install();
	 */
	function install() {
		global $db;
		
		if (!$this->_checkExistingSofortConstants($this->paymentMethod)) {
			return 'failed'; //word "failed" is needed by shopsystem!
		}
		
		//following defined constants are global(!) - that means, they are set only once and are used by most/all payment-methods
		if(!defined('MODULE_PAYMENT_SOFORT_MULTIPAY_APIKEY')) {
			
			$sofortStatuses = $this->_insertAndReturnSofortStatus();
			$tempStatus = (isset($sofortStatuses['temp'])&& !empty($sofortStatuses['temp']))? $sofortStatuses['temp'] : '';
			$abortedStatus = (isset($sofortStatuses['aborted'])&& !empty($sofortStatuses['aborted']))? $sofortStatuses['aborted'] : '';
			$unchangedStatus = (isset($sofortStatuses['unchanged'])&& !empty($sofortStatuses['unchanged']))? $sofortStatuses['unchanged'] : '';
			
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description, configuration_group_id, sort_order, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_APIKEY_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_APIKEY', '', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_APIKEY_DESC)."', '6', '2', now())");
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, set_function, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_AUTH_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_AUTH', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_KEYTEST_DEFAULT)."', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_AUTH_DESC)."', '6', '3', 'zen_cfg_select_option(array(),', now())");  //hide the input-field with an empty <select>
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, set_function, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_REASON_1_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_REASON_1', '-TRANSACTION-', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_REASON_1_DESC)."', '6', '8', 'zen_cfg_select_option(array(\'Kd-Nr. {{customer_id}}\',\'-TRANSACTION-\'), ', now())");
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_TEXT_REASON_2_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_TEXT_REASON_2', '" . HelperFunctions::escapeSql(STORE_NAME) . "', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_TEXT_REASON_2_DESC)."', '6', '9', now())");
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, set_function, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_IMAGE_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_IMAGE', 'Infographic', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_IMAGE_DESC)."', '6', '7', 'zen_cfg_select_option(array(\'Infographic\',\'Logo & Text\'), ', now())");
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, set_function, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_PROF_SETTINGS_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_PROF_SETTINGS', '', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_PROF_SETTINGS_DESC)."', '6', '20', 'zen_cfg_select_option(array(),', now())");  //hide the input-field with an empty <select>
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, set_function, use_function, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_TEMP_STATUS_ID_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_TEMP_STATUS_ID', '".HelperFunctions::escapeSql($tempStatus)."', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_TEMP_STATUS_ID_DESC)."', '6', '35', 'zen_cfg_pull_down_order_statuses(', 'zen_get_order_status_name', now())");
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, set_function, use_function, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_ABORTED_STATUS_ID_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_ABORTED_STATUS_ID', '".HelperFunctions::escapeSql($abortedStatus)."', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_ABORTED_STATUS_ID_DESC)."', '6', '35', 'zen_cfg_pull_down_order_statuses(', 'zen_get_order_status_name', now())");
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, set_function, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_LOG_ENABLED_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_LOG_ENABLED', 'False', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_LOG_ENABLED_DESC)."', '6', '28', 'zen_cfg_select_option(array(\'True\', \'False\'), ', now())");
			
			//only used while setting status - unchangeable and not shown in module-configuration
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_UNCHANGED_STATUS_ID_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_UNCHANGED_STATUS_ID', '".HelperFunctions::escapeSql($unchangedStatus)."', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_UNCHANGED_STATUS_ID_DESC)."', '6', '35', now())");
			
			$moduleVersion = HelperFunctions::getSofortmodulVersion();
			$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." (configuration_title, configuration_key, configuration_value, configuration_description,  configuration_group_id, sort_order, date_added)
				VALUES ('".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_MODULE_VERSION_TITLE)."', 'MODULE_PAYMENT_SOFORT_MULTIPAY_MODULE_VERSION', '".HelperFunctions::escapeSql($moduleVersion)."', '".HelperFunctions::escapeSql(MODULE_PAYMENT_SOFORT_MULTIPAY_MODULE_VERSION_DESC)."', '6', '2', now())");
			
			$this->_installSofortOrdersTable();
			$this->_installSofortOrdersNotificationTable();
			$this->_installSofortProductsTable();
		}
		return true;
	}
	
	
	/**
	 * checked ok
	 * removes all global entries in table 'configuration' (e.g.: configKey, reason etc.) if all multipay-modules are now uninstalled
	 * all payment-specific-data must be deleted in the module
	 * called by function remove() in module
	 */
	function remove() {
		global $db;
		
		$check_query = $db->Execute("SELECT * FROM ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." WHERE configuration_key like 'MODULE_PAYMENT_SOFORT_%_STATUS'");
		
		if ($check_query->RecordCount() === 0 ) {
			$db->Execute("DELETE FROM ".HelperFunctions::escapeSql(TABLE_CONFIGURATION)." WHERE configuration_key LIKE 'MODULE_PAYMENT_SOFORT_%'");
			
			//We don't want to delete data, but we could in some cases...
			//Notice: callback.php may throw warnings/errors, if they doesnt exist (in case of later notification)
			//$this->_removeSofortOrdersTable();
			//$this->_removeSofortOrdersNotificationTable();
			//$this->_removeSofortProductsTable();
		}
		return true;
	}
	
	
	/**
	 * checked ok
	 */
	function keys() {
		//unset, important for admin
		if (isset($_SESSION['sofort']['apiKeyIsValid'])) {
			unset($_SESSION['sofort']['apiKeyIsValid']);
		}
	}
	
	
	function _getGenderFromAddressBook($firstname, $lastname, $company, $streetAddress, $postcode, $city, $countryId, $zoneId) {
		global $db;
		$query = '	SELECT	entry_gender
					FROM	'.HelperFunctions::escapeSql(TABLE_ADDRESS_BOOK).'
					WHERE	entry_firstname		 = "'.HelperFunctions::escapeSql($firstname).'"
					AND		entry_lastname		 = "'.HelperFunctions::escapeSql($lastname).'"
					AND		entry_company		 = "'.HelperFunctions::escapeSql($company).'"
					AND		entry_street_address = "'.HelperFunctions::escapeSql($streetAddress).'"
					AND		entry_postcode		 = "'.HelperFunctions::escapeSql($postcode).'"
					AND		entry_city			 = "'.HelperFunctions::escapeSql($city).'"
					AND		entry_country_id	 = "'.HelperFunctions::escapeSql($countryId).'"
					AND		entry_zone_id		 = "'.HelperFunctions::escapeSql($zoneId).'" LIMIT	1';
		$values = $db->Execute($query);
		
		switch ($values->fields['entry_gender']) {
			case 'm': $salutation = 2;
				break;
			case 'f': $salutation = 3;
				break;
			default:  $salutation = '';
				break;
		}
		return $salutation;
	}
	
	
	/**
	 * checked ok
	 * check if order has temp-Status and buyer didnt finish the payment-process by "normal" way
	 * @return bool true (always)
	 */
	function _checkCancelOrder() {
		global $db;
		if (!empty($_SESSION['sofort']['cart_pn_sofortueberweisung_id'])) {
			$orderId = substr($_SESSION['sofort']['cart_pn_sofortueberweisung_id'], strpos($_SESSION['sofort']['cart_pn_sofortueberweisung_id'], '-') + 1);
			//$cartID = substr($_SESSION['sofort']['cart_pn_sofortueberweisung_id'], 0, strlen($_SESSION['cart']->cartID));
			$result = $db->Execute("SELECT orders_status FROM ".HelperFunctions::escapeSql(TABLE_ORDERS)." WHERE orders_id = '".(int) $orderId."' AND payment_module_code LIKE 'sofort_%'");
			$tempStatus = MODULE_PAYMENT_SOFORT_MULTIPAY_TEMP_STATUS_ID;
			if (HelperFunctions::statusIsUnchangedStatus($tempStatus)) {
				if (is_object($this->sofort)) {
					$this->sofort->logWarning("Warning: 'Temporary'-Status is set to 'Unchanged'-status. Don't use this status here and change in modulsettings! Check for 'abort'-status for this order could not be processed. Check order! Order-ID: $orderId");
				} elseif (is_object($this->invoice)) {
					$this->invoice->logWarning("Warning: 'Temporary'-Status is set to 'Unchanged'-status. Don't use this status here and change in modulsettings! Check for 'abort'-status for this order could not be processed. Check order! Order-ID: $orderId");
				}
				return true;
			}
			if ($result->fields['orders_status'] == $tempStatus) {
				unset($_SESSION['sofort']['cart_pn_sofortueberweisung_id']);
				$this->_cancelOrder( (int) $orderId, true);
			}
		}
		
		return true;
	}
	
	
	/**
	 * checked ok
	 * Check, if needed sofort-lang-constants exists.
	 * If not, include the needed lang-file (if doesnt exist: fallback is english-langfile)
	 * @return always true
	 */
	function _checkExistingSofortConstants($paymentMethod) {
		$paymentMethod = strtoupper($paymentMethod);
		
		$allowedPaymentMethods = array('SU', 'SL', 'LS', 'SR', 'SV');
		if(!in_array($paymentMethod, $allowedPaymentMethods)) {
			return false;
		}
		
		//security check - constant exists if lang-file exists
		if(defined('MODULE_PAYMENT_SOFORT_'.$paymentMethod.'_TEXT_TITLE')) {
			return true; //lang file was already included - nothing to do
		}
		
		$lngdir = DIR_FS_CATALOG_LANGUAGES;
		
		$iterator = new DirectoryIterator($lngdir);
		foreach ($iterator as $file){
			if ($file->isDir() && file_exists($lngdir.$file->getFilename().'/modules/payment/sofort_general.php')) $installedModulLangs[] = $file->getFilename();
		}
		
		//currently installed in this module
		if(in_array($_SESSION['language'], $installedModulLangs)) {
			include_once(DIR_FS_CATALOG_LANGUAGES.$_SESSION['language'].'/modules/payment/'.$this->code.'.php');
		} else {
			include_once(DIR_FS_CATALOG_LANGUAGES.'english/modules/payment/'.$this->code.'.php');
		}
		return true;
	}
	
	
	/**
	 * checked ok
	 * Insert all SOFORT-Status into table orders_status and return them
	 * @return array with all status
	 */
	function _insertAndReturnSofortStatus() {
		global $db;
		
		require_once('sofortInstall.php');
		
		//SOFORT-langs in this module
		$lngdir = DIR_FS_CATALOG_LANGUAGES;
		
		foreach (new DirectoryIterator($lngdir) as $file){
			if (file_exists($lngdir.$file->getFilename().'/modules/payment/sofort_general.php')) $sofortLangs[] = strtoupper($file->getFilename());
		}
		
		//current installed langs
		$installedLangs = array();
		$zenLangs = zen_get_languages();
		
		foreach ($zenLangs as $lang) {
			$installedLangs[$lang['id']] = strtoupper($lang['directory']);
		}
		
		//get the current highest orders_status_id
		$result = $db->Execute("SELECT MAX(orders_status_id) AS max_orders_status_id FROM ".HelperFunctions::escapeSql(TABLE_ORDERS_STATUS));
		$ordersStatusIdTemp = $result->fields['max_orders_status_id'] + 1;
		$ordersStatusIdConfirmed = $ordersStatusIdTemp + 1;
		$ordersStatusIdCanceled = $ordersStatusIdConfirmed + 1;
		$ordersStatusIdCheck = $ordersStatusIdCanceled + 1;
		$ordersStatusIdUnconfirmed = $ordersStatusIdCheck + 1;
		$ordersStatusIdInvoiceConfirmed = $ordersStatusIdUnconfirmed + 1;
		$ordersStatusIdSvUnpaid = $ordersStatusIdInvoiceConfirmed + 1;
		$ordersStatusIdAborted = $ordersStatusIdSvUnpaid + 1;
		$ordersStatusIdReturnedDebit = $ordersStatusIdAborted + 1;
		$ordersStatusIdRefunded = $ordersStatusIdReturnedDebit + 1;
		$ordersStatusIdDirectDebit = $ordersStatusIdRefunded + 1;
		$ordersStatusIdSvPaid = $ordersStatusIdDirectDebit + 1;
		$ordersStatusIdTranslateConfirmed = $ordersStatusIdSvPaid + 1;
		$ordersStatusIdUnchanged = $ordersStatusIdTranslateConfirmed + 1;
		
		$sofortStatuses = array();
		foreach($installedLangs as $installedLang) {
			
			//insert english for shop-languages, which are not included in this module
			if(in_array($installedLang, $sofortLangs)) {
				$sofortLang = $installedLang;
				$langId = array_search($sofortLang, $installedLangs);
			} else {
				$sofortLang = 'ENGLISH';
				$langId = array_search($installedLang, $installedLangs);
			}
			
			//insert temp-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('temp', $sofortLang);
			$sofortStatuses['temp'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['temp'] === false || $sofortStatuses['temp'] == '') {
				$sofortStatuses['temp'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdTemp, $langId);
			}
			
			//insert sv-unpaid-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('sv_unpaid', $sofortLang);
			$sofortStatuses['sv_unpaid'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['sv_unpaid'] === false || $sofortStatuses['sv_unpaid'] == '') {
				$sofortStatuses['sv_unpaid'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdSvUnpaid, $langId);
			}
			
			//insert sv-paid-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('sv_paid', $sofortLang);
			$sofortStatuses['sv_paid'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['sv_paid'] === false || $sofortStatuses['sv_paid'] == '') {
				$sofortStatuses['sv_paid'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdSvPaid, $langId);
			}
			
			//insert confirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('confirmed', $sofortLang);
			$sofortStatuses['confirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['confirmed'] === false || $sofortStatuses['confirmed'] == '') {
				$sofortStatuses['confirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdConfirmed, $langId);
			}
			
			//insert translate_confirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('translate_confirmed', $sofortLang);
			$sofortStatuses['translate_confirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['translate_confirmed'] === false || $sofortStatuses['translate_confirmed'] == '') {
				$sofortStatuses['translate_confirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdTranslateConfirmed, $langId);
			}
			
			//insert canceled-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('canceled', $sofortLang);
			$sofortStatuses['canceled'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['canceled'] === false || $sofortStatuses['canceled'] == '') {
				$sofortStatuses['canceled'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdCanceled, $langId);
			}
			
			//insert aborted-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('aborted', $sofortLang);
			$sofortStatuses['aborted'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['aborted'] === false || $sofortStatuses['aborted'] == '') {
				$sofortStatuses['aborted'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdAborted, $langId);
			}
			
			//insert check-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('check', $sofortLang);
			$sofortStatuses['check'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['check'] === false || $sofortStatuses['check'] == '') {
				$sofortStatuses['check'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdCheck, $langId);
			}
			
			//insert unconfirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('unconfirmed', $sofortLang);
			$sofortStatuses['unconfirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['unconfirmed'] === false || $sofortStatuses['unconfirmed'] == '') {
				$sofortStatuses['unconfirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdUnconfirmed, $langId);
			}
			
			//insert invoice-confirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('invoice_confirmed', $sofortLang);
			$sofortStatuses['invoice_confirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['invoice_confirmed'] === false || $sofortStatuses['invoice_confirmed'] == '') {
				$sofortStatuses['invoice_confirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdInvoiceConfirmed, $langId);
			}
			
			//insert returned-debit-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('returned_debit', $sofortLang);
			$sofortStatuses['returned_debit'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['returned_debit'] === false || $sofortStatuses['returned_debit'] == '') {
				$sofortStatuses['returned_debit'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdReturnedDebit, $langId);
			}
			
			//insert refunded-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('refunded', $sofortLang);
			$sofortStatuses['refunded'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['refunded'] === false || $sofortStatuses['refunded'] == '') {
				$sofortStatuses['refunded'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdRefunded, $langId);
			}
			
			//insert direct-debit-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('direct_debit', $sofortLang);
			$sofortStatuses['direct_debit'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['direct_debit'] === false || $sofortStatuses['direct_debit'] == '') {
				$sofortStatuses['direct_debit'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdDirectDebit, $langId);
			}
			
			//insert unchanged-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('unchanged', $sofortLang);
			$sofortStatuses['unchanged'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['unchanged'] === false || $sofortStatuses['unchanged'] == '') {
				$sofortStatuses['unchanged'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdUnchanged, $langId);
			}
		}
		return  $sofortStatuses;
	}
	
	
	/**
	 * checked ok
	 */
	function _getStatusIdIfExistInDb($newOrdersStatusName, $langId) {
		global $db;
		
		if(!$newOrdersStatusName) return false;
		
		$queryResult = $db->Execute('SELECT orders_status_id
				FROM '.HelperFunctions::escapeSql(TABLE_ORDERS_STATUS).' 
				WHERE language_id = "'.HelperFunctions::escapeSql($langId).'"
				AND orders_status_name = "'.HelperFunctions::escapeSql($newOrdersStatusName).'"
				LIMIT 1');
		
		if ($queryResult->RecordCount() < 1) {
			return false;
		}else{
			//if string exists, the orders_status_id is returned
			return $queryResult->fields['orders_status_id'];
		}
	}
	
	
	/**
	 * checked ok
	 * insert given statusstring into DB - empty strings will not be inserted
	 * @return $orders_status_id from DB OR false
	 */
	function _insertStatusInDb($newOrdersStatusName, $ordersStatusId, $langId) {
		global $db;
		if (!$newOrdersStatusName) return false;
		$db->Execute("INSERT INTO ".HelperFunctions::escapeSql(TABLE_ORDERS_STATUS)." (orders_status_id, language_id, orders_status_name)
			values ('".HelperFunctions::escapeSql($ordersStatusId)."', '".HelperFunctions::escapeSql($langId)."', '".HelperFunctions::escapeSql($newOrdersStatusName)."')");
		return $ordersStatusId;
	}
	
	
	/**
	 * checked ok
	 * returns the statusname for the given $status and $lang
	 * return max the first 32 chars! (because db-field = varchar(32) )
	 */
	function _getNewOrdersStatusName($status, $lang) {
		switch ($status) {
			case 'temp': 				$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_TEMP_'.$lang); break;
			case 'sv_unpaid':			$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_SV_UNPAID_'.$lang); break;
			case 'sv_paid':				$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_SV_PAID_'.$lang); break;
			case 'confirmed':			$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_CONFIRMED_'.$lang); break;
			case 'translate_confirmed': $newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_CONFIRMED_2_'.$lang); break;
			case 'invoice_confirmed':	$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_INVOICE_CONFIRMED_'.$lang); break;
			case 'unconfirmed':			$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_UNCONFIRMED_'.$lang); break;
			case 'canceled':			$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_CANCELED_'.$lang); break;
			case 'aborted':				$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_ABORTED_'.$lang); break;
			case 'check':				$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_CHECK_'.$lang); break;
			case 'returned_debit':		$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_RETURNED_DEBIT_'.$lang); break;
			case 'refunded':			$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_REFUNDED_'.$lang); break;
			case 'direct_debit':		$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_DIRECT_DEBIT_'.$lang); break;
			case 'unchanged':			$newOrdersStatusName = constant('MODULE_PAYMENT_SOFORT_MULTIPAY_STATE_UNCHANGED_'.$lang); break;
		}
		// if string is not cut to 32 chars, status will be reinserted with every installation
		$newOrdersStatusName = substr($newOrdersStatusName, 0, 32);

		return $newOrdersStatusName;
	}
	
	
	/**
	 * checked ok
	 */
	function _installSofortOrdersTable() {
		global $db;
		$sql = '
			CREATE TABLE IF NOT EXISTS `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders` (
				`id` int(11) unsigned NOT NULL AUTO_INCREMENT,
				`orders_id` int(11) unsigned NOT NULL,
				`transaction_id` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`payment_method` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`payment_secret` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`date_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
				PRIMARY KEY (`id`)
			)
		';
		$db->Execute($sql);
	}
	
	
	/**
	 * checked ok
	 */
	function _removeSofortOrdersTable() {
		global $db;
		$sql = 'DROP TABLE `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders`';
		$db->Execute($sql);
	}
	
	
	/**
	 * checked ok
	 */
	function _installSofortOrdersNotificationTable() {
		global $db;
		$sql = '
			CREATE TABLE IF NOT EXISTS `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders_notification` (
				`id` int(11) unsigned NOT NULL AUTO_INCREMENT,
				`sofort_orders_id` int(11) unsigned NOT NULL,
				`items` text COLLATE utf8_unicode_ci NOT NULL,
				`amount` float NOT NULL,
				`customer_comment` text COLLATE utf8_unicode_ci NOT NULL,
				`seller_comment` text COLLATE utf8_unicode_ci NOT NULL,
				`status_id` int(11) unsigned NOT NULL,
				`status` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`status_reason` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`invoice_status` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`invoice_objection` varchar(45) COLLATE utf8_unicode_ci NOT NULL,
				`date_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
				PRIMARY KEY (`id`)
			)
		';
		$db->Execute($sql);
	}
	
	
	/**
	 * checked ok
	 */
	function _removeSofortOrdersNotificationTable() {
		global $db;
		$sql = 'DROP TABLE `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_orders_notification`';
		$db->Execute($sql);
	}
	
	
	/**
	 * checked ok
	 */
	function _installSofortProductsTable() {
		global $db;
		$sql = '
			CREATE TABLE IF NOT EXISTS `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_products` (
				`id` int(11) unsigned NOT NULL AUTO_INCREMENT,
				`orders_id` int(11) unsigned NOT NULL,
				`orders_products_id` int(11) unsigned NOT NULL,
				`item_id` text COLLATE utf8_unicode_ci NOT NULL,
				PRIMARY KEY (`id`)
			)
		';
		$db->Execute($sql);
	}
	
	
	/**
	 * checked ok
	 */
	function _removeSofortProductsTable() {
		global $db;
		$sql = 'DROP TABLE `'.HelperFunctions::escapeSql(DB_PREFIX).'sofort_products`';
		$db->Execute($sql);
	}
	
	
	/**
	 * checked ok
	 * get shop order total
	 * @return float $shopEndprice
	 */
	function _getShopTotal($orderTotals, $order) {
		//Frequent sources of errors: shipping-tax, external modules, sort order of shown 'ot_'-modules
		$ot_totalTotal = 0;
		
		if (MODULE_ORDER_TOTAL_INSTALLED) {
			foreach ($orderTotals as $oneTotal) {
				if ($oneTotal['code'] == 'ot_total') {
					$ot_totalTotal = $oneTotal['value'];
				}
			}
		}
		
		$orderObjectTotal = $order->info['total'];
		
		//use the higher one
		$shopTotal = 0;
		
		if ($ot_totalTotal >= $orderObjectTotal) {
			$shopTotal = $ot_totalTotal;
		} else if ($orderObjectTotal > $ot_totalTotal) {
			$shopTotal = $orderObjectTotal;
		}
		
		//if payment is not EUR: calculate with exchange rate
		$shopTotal = $shopTotal * $order->info['currency_value'];
		
		return number_format($shopTotal, 2, '.','');
	}
	
	
	/**
	 * checked ok
	 * returns currently given value - activate code if needed
	 * @param string $reason like 'K?ufer A. M?ller'
	 * @return string - converted String
	 */
	function _convertReason($reason) {
		return $reason;
		
		//activate if needed...
		
		//replace special chars in $reason e.g. ?=>ae, ?=>ss incl. foreign non-ascii-signs etc.
		//$oldLocale = setlocale(LC_ALL, 0);
		//setlocale(LC_ALL, "de_DE.utf8");
		//$shopEncoding = HelperFunctions::getIniValue('shopEncoding');
		//$encoding = mb_detect_encoding('aaa'.$reason, "ISO-8859-1, UTF-8");
		//$convertedReason = substr(iconv($shopEncoding, "ASCII//TRANSLIT", 'aaa'.$reason),3);
		//setlocale(LC_ALL, $oldLocale);
		//return $convertedReason;
	}
	
	
	/**
	 * checked ok
	 * get the reasons for the given payment method
	 * @return array with reason1 and reason2
	 */
	function _getReasons($paymentMethod, $customerId, $order, $orderId) {
		
		if ($paymentMethod == 'SV') {  //SV has only one reason
			$reason_1 = str_replace('{{order_id}}', $orderId, MODULE_PAYMENT_SOFORT_SV_REASON_2);
			$reason_1 = str_replace('{{customer_id}}', $customerId, $reason_1);
			$reason_1 = str_replace('{{transaction_id}}', '-TRANSACTION-', $reason_1);
			$reason_1 = $this->_convertReason($reason_1);
			$reason_1 = substr($reason_1, 0, 27);
			
			$reason_2 = ''; //SV has only one reason, the 2nd is set by SOFORT
		} else {
			$reason_1 = str_replace('{{order_id}}', $orderId, MODULE_PAYMENT_SOFORT_MULTIPAY_REASON_1);
			$reason_1 = str_replace('{{customer_id}}', $customerId, $reason_1);
			$reason_1 = str_replace('{{transaction_id}}', '-TRANSACTION-', $reason_1);
			$reason_1 = $this->_convertReason($reason_1);
			$reason_1 = substr($reason_1, 0, 27);
			
			$reason_2 = str_replace('{{order_id}}', $orderId, MODULE_PAYMENT_SOFORT_MULTIPAY_TEXT_REASON_2);
			$reason_2 = str_replace('{{customer_id}}', $customerId, $reason_2);
			$reason_2 = str_replace('{{order_date}}', strftime(DATE_FORMAT_SHORT), $reason_2);
			$reason_2 = str_replace('{{customer_name}}', $order->customer['firstname'] . ' ' . $order->customer['lastname'], $reason_2);
			$reason_2 = str_replace('{{customer_company}}', $order->customer['company'], $reason_2);
			$reason_2 = str_replace('{{customer_email}}', $order->customer['email_address'], $reason_2);
			$reason_2 = str_replace('{{transaction_id}}', '-TRANSACTION-', $reason_2);
			$reason_2 = $this->_convertReason($reason_2);
			$reason_2 = substr($reason_2, 0, 27);
		}
		
		$pattern = '/[^\sa-zA-Z0-9+,-. ]/';
		$replace = ' ';
		$reason_1 = preg_replace($pattern, $replace, $reason_1);
		$reason_2 = preg_replace($pattern, $replace, $reason_2);
		
		$reasons = array();
		$reasons[0] = $reason_1;
		$reasons[1] = $reason_2;
		return $reasons;
		
	}
	
	
	/**
	 * checked ok
	 * delete all data of this buying-proccess - currently only needed for SV;
	 * @see /includes/modules/pages/checkout_process/header_php.php
	 * @return bool true (always)
	 */
	function _resetCartAndDeleteSessionData() {
		global $order_total_modules, $zco_notifier;
		
		if (is_object($_SESSION['cart'])) {
			$_SESSION['cart']->reset(true);
		}
		
		// unregister session variables used during checkout
		unset($_SESSION['sendto']);
		unset($_SESSION['billto']);
		unset($_SESSION['shipping']);
		unset($_SESSION['payment']);
		unset($_SESSION['comments']);
		if (is_object($order_total_modules)) {
			$order_total_modules->clear_posts();//ICW ADDED FOR CREDIT CLASS SYSTEM
		}
		
		// This should be before the zen_redirect:
		if (is_object($zco_notifier)) {
			$zco_notifier->notify('NOTIFY_HEADER_END_CHECKOUT_PROCESS');
		}
		
		return true;
	}
	
	
	/**
	 * checked ok
	 * @see zen_remove_order() in /admin/includes/functions/general.php
	 * @param int $orderId
	 * @param bool $restock
	 */
	function _cancelOrder($orderId, $restock = false) {
		global $db;
		//if buyer will not successfully abort the payment or close the SOFORT-Wizard and the goes to the
		//shop-index.php and want to pay again -> set the orderStatus to "aborted"
		
		if ($restock == true) {
			$order = $db->Execute("SELECT products_id, products_quantity
				FROM ".HelperFunctions::escapeSql(TABLE_ORDERS_PRODUCTS)."
				WHERE orders_id = '".(int)$orderId."'");
			
			while (!$order->EOF) {
				$db->Execute("UPDATE ".HelperFunctions::escapeSql(TABLE_PRODUCTS)."
					SET products_quantity = products_quantity + ".(int)$order->fields['products_quantity'].",
						products_ordered = products_ordered - ".(int)$order->fields['products_quantity']."
					WHERE products_id = '".(int)$order->fields['products_id']."'");
				$order->MoveNext();
			}
		}
		
		//update status and customer-history
		$time = date("d.m.Y, G:i:s");
		
		HelperFunctions::updateTableOrdersStatus($orderId, MODULE_PAYMENT_SOFORT_MULTIPAY_ABORTED_STATUS_ID);
		
		HelperFunctions::insertHistoryEntry($orderId, MODULE_PAYMENT_SOFORT_MULTIPAY_ABORTED_STATUS_ID, MODULE_PAYMENT_SOFORT_MULTIPAY_ORDER_CANCELED, '', $time);
	}
	
	
	/**
	 * checked ok
	 * check if order has status "paid" and insert "paid"-status into table_orders and sofort_orders and set comments to history
	 * (sets log-entries in all currious situations)
	 * @return bool TRUE (if errors occured: FALSE)
	 * @todo this function is too long and should be splittet - Notice: this function is the most complicated function of this file!!!
	 */
	function _insertPaidStatus($orderId, $transactionId) {
		//create transactionData-Object
		$SofortLib_TransactionData = new SofortLib_TransactionData(MODULE_PAYMENT_SOFORT_MULTIPAY_APIKEY);
		if (!is_object($SofortLib_TransactionData) || !($SofortLib_TransactionData instanceof SofortLib_TransactionData)) {
			if (is_object($this->invoice)) {
				$this->invoice->logError("TransactionData-Object corrupt. Maybe status could not be correct set! Order-ID: $orderId. Check status of order!");
			} else {
				$this->sofort->logError("TransactionData-Object corrupt. Maybe status could not be correct set! Order-ID: $orderId. Check status of order!");
			}
			return false;
		}
		//get transactionID from table orders
		$shopTransactionId = HelperFunctions::getTransactionIdFromShop($orderId);
		if (!$shopTransactionId || $shopTransactionId != $transactionId){
			$SofortLib_TransactionData->logError("Transaction-ID not found or corrupt for order-ID: $orderId. Maybe status could not be correct set! Check status of order!");
			return false;
		}
		
		
		//fetch all transactionData from SOFORT
		$SofortLib_TransactionData->setTransaction($transactionId);
		$SofortLib_TransactionData->sendRequest();
		if ($SofortLib_TransactionData->isError()) {
			$SofortLib_TransactionData->logError("Error in TransactionData: ".print_r($SofortLib_TransactionData->getError(), true)."Transaction-ID: $transactionId.
				Order-ID: $orderId. Maybe status could not be correct set! Check status of order!");
			return false;
		}
		
		//check if SOFORT-paymentSecret is equal to shop-paymentSecret
		$paymentSecret = $SofortLib_TransactionData->getUserVariable(2); //paymentSecret was sent to SOFORT in the 3rd user variable
		if (!HelperFunctions::checkPaymentSecret($paymentSecret, $transactionId)) {
			$SofortLib_TransactionData->logError("Check of Payment-Secret failed for given Transaction-ID: $transactionId! Checked paymentSecret: $paymentSecret. Order-ID: $orderId. Check order!");
			return false;
		}
		
		//check if SOFORT-orderId is equal to shop-orderId
		$shopOrderId = HelperFunctions::getOrderId($transactionId);
		$transDataOrderId = $SofortLib_TransactionData->getUserVariable(0);
		if (!$shopOrderId || $shopOrderId != $orderId || $orderId != $transDataOrderId) {
			$SofortLib_TransactionData->logError("Found order-IDs are not equal ($orderId/$shopOrderId/$transDataOrderId) for Transaction-ID: $transactionId! Check order!");
			return false;
		}
		
		$time = date('d.m.Y, G:i:s', strtotime($SofortLib_TransactionData->getTime()));
		$statusReason = $SofortLib_TransactionData->getStatusReason();
		$status = $SofortLib_TransactionData->getStatus();
		$paymentMethod = $SofortLib_TransactionData->getPaymentMethod();
		
		$PnagInvoice = '';
		if ($paymentMethod == 'sr') {
			$PnagInvoice = new PnagInvoice(MODULE_PAYMENT_SOFORT_MULTIPAY_APIKEY, $transactionId);
			if ($PnagInvoice->isError()) {
				$SofortLib_TransactionData->logError("Errors while creating invoice-object: ".
					print_r($PnagInvoice->getErrors(), true)." Transaction-ID: $transactionId, Order-ID: $orderId. Check order!");
				return false;
			}
		}
		
		$dbEntries = array(
				'amount' => $SofortLib_TransactionData->getAmount(),
				'status' => $status,
				'status_reason' => $statusReason,
		);
		
		if ($paymentMethod == 'sr') {
			$dbEntries['status_id'] = $PnagInvoice->getState();
		}
		
		$statusShort = HelperFunctions::getShortStatus($paymentMethod, $status, $statusReason);
		$newShopStatus = HelperFunctions::getNewShopStatusId($statusShort, $orderId, $SofortLib_TransactionData);
		$historyComments = HelperFunctions::getHistoryComments($statusShort, $orderId, $SofortLib_TransactionData, $transactionId, false, $SofortLib_TransactionData);
		
		//check if notification was faster than call of successURL
		if (HelperFunctions::hasEntriesInTableSofortNotifications($transactionId)) {
			$SofortLib_TransactionData->log("Status of first notification was already set by notification (not by success-URL) for order-ID: $orderId. There is nothing to do here.");
			return true; //everything was done by notification
		}
		
		//save all relevant data in table sofort_notification
		//this DB-entry also prevents race-conditions between call of successUrl and notification
		//@see hasEntriesInTableSofortNotifications() and statusAndReasonDidNotChange()
		//must be set (preferably direct) after call of getHistoryComments() (because of a checks there inside)
		HelperFunctions::insertSofortOrdersNotification($transactionId, $dbEntries, $SofortLib_TransactionData);
		
		//special features for SR
		if ($paymentMethod == 'sr') {
		
			//if buyer has changed addresses at sofort-wizard -> set new addresses also in shopsystem
			if ($statusShort == 'sr_pen_con_inv') {
				HelperFunctions::updateShopAdresses($SofortLib_TransactionData->getInvoiceAddress(), $SofortLib_TransactionData->getShippingAddress(), $orderId, $SofortLib_TransactionData);
			}
		}
		
		//insert all data into orders-table and orders_status_history-table
		if ($historyComments['buyer']) {
			if (!$newShopStatus) {
				$SofortLib_TransactionData->logError("Status for buyer-comment (".$historyComments['buyer'].") unknown. Comment will not be
					inserted. Check order! Order-Id: $orderId, Transaction-ID: $transactionId!");
				return false;
			}
			HelperFunctions::insertHistoryEntry($orderId, $newShopStatus, $historyComments['buyer'], '', $time);
		}
		
		if ($historyComments['seller']) {
			HelperFunctions::insertHistoryEntry($orderId, -1, $historyComments['seller'], '', $time);
		}
		
		//if $newShopStatus is set, set it into table orders
		if ($newShopStatus) {
			HelperFunctions::updateTableOrdersStatus($orderId, $newShopStatus);
		}
		
		return true;
	}
	
	
	/**
	 * checked ok
	 * checks, if current payment-method is installed (installed != enabled)
	 */
	function _isInstalled() {
		$installedPaymentMethods = explode(';', MODULE_PAYMENT_INSTALLED);
		if (in_array($this->code.'.php', $installedPaymentMethods)) {
			return true;
		} else {
			return false;
		}
	}
	
	
	/**
	 * save bankdata in users history and return link to sv-bankdata-page
	 * checked ok
	 * @return string - link to sv-bankdata-page
	 */
	function _insertSvBankdataAndGetLinkToBankdataPage($orderId) {
		//save sofortvorkasse-bankdata in customer history and show bankdata-page
		$bankdata = MODULE_PAYMENT_SOFORT_SV_CHECKOUT_HOLDER_TEXT.		  ' '.HelperFunctions::htmlMask($_GET['holder'])		.' -- '.
					MODULE_PAYMENT_SOFORT_SV_CHECKOUT_ACCOUNT_NUMBER_TEXT.' '.HelperFunctions::htmlMask($_GET['account_number']).' -- '.
					MODULE_PAYMENT_SOFORT_SV_CHECKOUT_IBAN_TEXT.		  ' '.HelperFunctions::htmlMask($_GET['iban'])			.' -- '.
					MODULE_PAYMENT_SOFORT_SV_CHECKOUT_BANK_CODE_TEXT.	  ' '.HelperFunctions::htmlMask($_GET['bank_code'])		.' -- '.
					MODULE_PAYMENT_SOFORT_SV_CHECKOUT_BIC_TEXT.			  ' '.HelperFunctions::htmlMask($_GET['bic'])			.' -- '.
					MODULE_PAYMENT_SOFORT_SV_CHECKOUT_AMOUNT_TEXT.		  ' '.HelperFunctions::htmlMask($_GET['amount'])		.' -- '.
					MODULE_PAYMENT_SOFORT_SV_CHECKOUT_REASON_1_TEXT.	  ' '.HelperFunctions::htmlMask($_GET['reason_1'])		.' -- '.
					MODULE_PAYMENT_SOFORT_SV_CHECKOUT_REASON_2_TEXT.	  ' '.HelperFunctions::htmlMask($_GET['reason_2']);
		
		$lastOrderStatus = HelperFunctions::getLastOrderStatus($orderId);
		//only for security - should normally not happen
		if (!$lastOrderStatus) {
			$lastOrderStatus = DEFAULT_ORDERS_STATUS_ID;
		}
		
		$time = date("d.m.Y, G:i:s");
		if (SEND_EMAILS == 'true') {
			HelperFunctions::insertHistoryEntry($orderId, $lastOrderStatus, $bankdata, 1, $time);
		} else {
			HelperFunctions::insertHistoryEntry($orderId, $lastOrderStatus, $bankdata, '', $time);
		}
		
		//show bankdata to buyer when paying with vorkasse
		$get = '&holder='.HelperFunctions::htmlMask($_GET['holder']).'&account_number='.HelperFunctions::htmlMask($_GET['account_number']).
			'&iban='.HelperFunctions::htmlMask($_GET['iban']).'&bank_code='.HelperFunctions::htmlMask($_GET['bank_code']).
			'&bic='.HelperFunctions::htmlMask($_GET['bic']).'&amount='.HelperFunctions::htmlMask($_GET['amount']).
			'&reason_1='.HelperFunctions::htmlMask($_GET['reason_1']).'&reason_2='.HelperFunctions::htmlMask($_GET['reason_2']);
		return zen_href_link('sofort_vorkasse_confirm', $get, 'SSL', true, false);
	}
	
	
	/**
	 * send sv-bankdata to customer and seller (if enabled)
	 * @return true (always)
	 * checked ok
	 */
	function _sendSvBankDataEmail() {
		global $order;
		//send email only if its enabled by seller and if data for sending exist
		if (SEND_EMAILS == 'true' && isset($_GET['holder']) && isset($_GET['amount'])) {
			$paymentBlock  = "\n\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_HOLDER_TEXT." ".HelperFunctions::htmlMask($_GET['holder']);
			$paymentBlock .= "\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_ACCOUNT_NUMBER_TEXT." ".HelperFunctions::htmlMask($_GET['account_number']);
			$paymentBlock .= "\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_IBAN_TEXT." ".HelperFunctions::htmlMask($_GET['iban']);
			$paymentBlock .= "\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_BANK_CODE_TEXT." ".HelperFunctions::htmlMask($_GET['bank_code']);
			$paymentBlock .= "\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_BIC_TEXT." ".HelperFunctions::htmlMask($_GET['bic']);
			$paymentBlock .= "\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_AMOUNT_TEXT." ".number_format(HelperFunctions::htmlMask($_GET['amount']),2,',','.'). ' Euro;';
			$paymentBlock .= "\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_REASON_1_TEXT." ".HelperFunctions::htmlMask($_GET['reason_1']);
			$paymentBlock .= "\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_REASON_1_TEXT." ".HelperFunctions::htmlMask($_GET['reason_2']);
			$paymentBlock .= "\n\n".MODULE_PAYMENT_SOFORT_SV_CHECKOUT_REASON_HINT."\n\n";
						
			$customerEmail = $order->customer['email_address'];
			$customerName = $order->customer['name'];
			$mailSubject = MODULE_PAYMENT_SOFORT_GENERAL_EMAIL_SUBJECT;
			$mailText = MODULE_PAYMENT_SOFORT_GENERAL_EMAIL_TEXT;
			$mailText = str_replace('{{payment_block}}', $paymentBlock, $mailText);
			$mailText .= "\n\n".STORE_NAME;
			
			//send mail
			zen_mail($customerName, $customerEmail, $mailSubject, $mailText, STORE_NAME, EMAIL_FROM);
			if (SEND_EXTRA_ORDER_EMAILS_TO != '') {
				zen_mail($customerName, SEND_EXTRA_ORDER_EMAILS_TO, $mailSubject, $mailText, STORE_NAME, EMAIL_FROM);
			}
		}
		
		return true;
	}
}